"""This script is run by meson.build.

It *directly* imports the built .so file and generates a stub file for it,
circumventing any calls to `import _pymmcore_nano` in the process.
"""

import importlib.util
import re
import subprocess
import sys
from pathlib import Path
from types import ModuleType

from nanobind.stubgen import StubGen


def load_module_from_filepath(name: str, filepath: str) -> ModuleType:
    # Create a module spec
    spec = importlib.util.spec_from_file_location(name, filepath)
    if spec is None or spec.loader is None:
        raise ImportError(f"Cannot load module '{name}' from '{filepath}'")
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    sys.modules[name] = module
    return module


def build_stub(module_path: Path, output_path: str):
    module_name = module_path.stem.split(".")[0]
    module = load_module_from_filepath(module_name, str(module_path))
    s = StubGen(module, include_docstrings=True, include_private=False)
    s.put(module)
    dest = Path(output_path)
    stub_txt = s.get()

    # HACKY FIXES FOR BAD TYPE HINTS ... try to fix my in _pymmcore_nano.cc instead
    # fix a couple types caused by our own lambda functions
    stub_txt = re.sub(r'"std::__1::tuple<([^>]+)>"', r"tuple[\1]", stub_txt)
    stub_txt = re.sub("double", "float", stub_txt)

    # fix nanobind CapsuleType until we do better with numpy arrays
    stub_txt = "from typing import Any\n" + stub_txt
    stub_txt = re.sub("types.CapsuleType", "Any", stub_txt)

    # remove extra newlines and let ruff-format add them back
    stub_txt = re.sub("\n\n", "\n", stub_txt)

    dest.write_text(stub_txt)
    ruff = Path(sys.executable).parent / "ruff"
    _ruff = str(ruff) if ruff.exists() else "ruff"
    subprocess.run([_ruff, "format", output_path], check=True)
    subprocess.run(
        [
            _ruff,
            "check",
            "--fix-only",
            "--unsafe-fixes",
            "--ignore=D",
            output_path,
        ]
    )


if __name__ == "__main__":
    if len(sys.argv) > 1:
        module_path = Path(sys.argv[1])
    else:
        build_dir = Path(__file__).parent.parent / "builddir"
        module_path = next(x for x in build_dir.glob("_pymmcore_nano.*") if x.is_file())

    if len(sys.argv) > 2:
        output = Path(sys.argv[2])
    else:
        source_dir = Path(__file__).parent.parent / "src"
        output = source_dir / "pymmcore_nano" / "_pymmcore_nano.pyi"

    build_stub(module_path, str(output))



import os
import pathlib
import shutil


def main():
    dist_dir = pathlib.Path(os.environ["MESON_DIST_ROOT"]).resolve()

    for f in (dist_dir / "src" / "mmCoreAndDevices").iterdir():
        if f.name not in ("MMCore", "MMDevice"):
            if f.is_dir():
                shutil.rmtree(f)
            else:
                os.remove(f)

    shutil.rmtree(dist_dir / "tests")


if __name__ == "__main__":
    main()



"""Pull in the latest versions of the source files from mmCoreAndDevices."""

import shutil
import subprocess
import sys
import tempfile
from itertools import chain
from pathlib import Path

REPO_URL = "https://github.com/micro-manager/mmCoreAndDevices.git"
TARGET = Path(__file__).resolve().parent.parent / "src" / "mmCoreAndDevices"
FOLDERS_TO_SYNC = [
    "MMCore",
    "MMDevice",
    "DeviceAdapters/DemoCamera",
    "DeviceAdapters/SequenceTester",
    "DeviceAdapters/NotificationTester",
    "DeviceAdapters/Utilities",
]
REMOVE_GLOBS = ["*.am", "*.vcxproj*"]  # files to remove after sync


def sync_directories(src: Path, dest: Path) -> None:
    """Overwrite destination directory with source directory."""
    if dest.exists():
        shutil.rmtree(dest)
    shutil.copytree(src, dest)

    # recursively remove files with a specific glob pattern
    for pattern in REMOVE_GLOBS:
        for p in Path(dest).rglob(pattern):
            p.unlink()


def main(branch: str = "main") -> None:
    existing_meson_files = Path(TARGET).rglob("meson.build")
    existing_meson_wraps = Path(TARGET).rglob("*.wrap")

    # Clone repository to a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        tmp_path = Path(temp_dir)

        # copy existing meson files to the temp directory
        try:
            print(f"Cloning {REPO_URL} into {temp_dir}...")
            cmd = [
                "git",
                "clone",
                "--branch",
                branch,
                "--depth",
                "1",
                REPO_URL,
                temp_dir,
                "--single-branch",
            ]
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Error during git operation: {e}")
            return

        # remember our own meson files
        meson_tmp = tmp_path / "meson_files"
        meson_tmp.mkdir(exist_ok=True)  # store the original meson files here
        to_restore: list[tuple[Path, Path]] = []
        for i, file in enumerate(chain(existing_meson_files, existing_meson_wraps)):
            stored = Path(shutil.copy2(file, meson_tmp / f"{file.name}_{i}"))
            to_restore.append((file, stored))

        # Sync specified folders
        try:
            for folder in FOLDERS_TO_SYNC:
                src_path = tmp_path / folder
                dest_path = TARGET / folder
                print(f"Syncing {src_path} to {dest_path}...")
                sync_directories(src_path, dest_path)

            print("Sync complete.")
        except Exception as e:
            print(f"Unexpected error: {e}")

        # restore the original meson files
        for original, stored in to_restore:
            original.parent.mkdir(parents=True, exist_ok=True)
            shutil.move(stored, original)


if __name__ == "__main__":
    if len(sys.argv) > 1:
        branch_name = sys.argv[1]
    else:
        branch_name = "main"

    main(branch_name)
